#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <errno.h>
#include "basic.h"
#include "treiber_tagged.h"

static int dbg = 0;


/*  Nodes */

/* Real stack node */

typedef struct stack_node_t {  
  void *val ;
  struct stack_node_t *next ;
} stack_node_t ;

/* Heap nodes, we write our own allocator! */

#define LOGSZ 16
#define MAXNODES (1U << LOGSZ)
#define MASK (MAXNODES-1)

typedef struct heap_node_t {
  stack_node_t node ;
  struct heap_node_t *next ; // For free list(s)
} heap_node_t ;

static pthread_mutex_t* heap_mutex ;
static heap_node_t heap[MAXNODES] ;
static heap_node_t *free_list = NULL;
/* static  __thread heap_node_t *local_free_list = NULL; */

void init_heap(void) {
  heap_mutex = alloc_mutex() ;
  for (int k = 1 ; k < MAXNODES-1 ; k++) {
    heap[k].next = &heap[k+1] ;
  }
  heap[MAXNODES-1].next = NULL ;
  free_list = &heap[1] ;
}

stack_node_t *alloc_stack_node(void *v) {
  heap_node_t *r ;
  for ( ; ; ) {
    lock_mutex(heap_mutex) ;
    while (free_list == NULL) {
      unlock_mutex(heap_mutex) ;
      delay(50) ;
      lock_mutex(heap_mutex) ;
    }
    r = free_list ;
    free_list = free_list->next ;
    if (dbg > 1) {
      fprintf(stderr,"ALLOC 0x%lX\n",r-heap) ;
    }
    unlock_mutex(heap_mutex) ;
    stack_node_t *p = &r->node ;
    if ((void *)p != (void *)r) {
      fprintf(stderr,"assert false\n") ;
      exit(2) ;
    }
    p->val = v ;
    p->next = NULL ;
    return p ;
  }
}

void free_stack_node(stack_node_t *p) {
  lock_mutex(heap_mutex) ;
  heap_node_t *q = (heap_node_t *)p ;
  q->next = free_list ;
  free_list = q ;
  if (dbg > 1) {
    fprintf(stderr,"FREE 0x%lX\n",free_list-heap) ;
  }
  unlock_mutex(heap_mutex) ;
}


static unsigned low16(uintmax_t x) {
  return x & MASK ;
}

static unsigned high16(uintmax_t x) {
  return x >> LOGSZ ;
}


static stack_node_t *getptr(tagged_t p) {
  int low =low16(p) ;
  if (low == 0) return NULL ;
  return &(heap[low].node) ;
}


static unsigned gettag(tagged_t p) {
  return high16(p) ;
}

static tagged_t pack(stack_node_t *p, unsigned tag) {
  heap_node_t *q = (heap_node_t *)p ;
  if (q == NULL) q = heap ;
  unsigned idx = q-heap ;
  return idx | (tag << LOGSZ) ;
}
/* Stack proper */

#define CAS(x,y,z) __sync_bool_compare_and_swap(x,y,z)

/* Stack */

stack_t *alloc_stack(int nprocs) {
  stack_t *r =  malloc_check(sizeof(*r)) ;
  r->top = 0 ;
  return r ;
}

void pp_stack(stack_t *p,void pp(void *p)) {
  lock_mutex(heap_mutex) ;
  fprintf(stderr,"0x%" PRIXMAX ":",p->top) ;
  stack_node_t *q = getptr(p->top) ;
  while (q != NULL) {
    fprintf(stderr," ") ; pp(q->val) ;
    q = q->next ;
  }
  fprintf(stderr,"\n") ;
  unlock_mutex(heap_mutex) ;
}

void free_stack(stack_t *p) { free(p) ; }


void push(stack_t *p, void *v) {
  stack_node_t *q = alloc_stack_node(v) ;
  tagged_t old,q_packed ;
  do {
    SYNC() ;
    old = p->top ;
    stack_node_t *realold = getptr(old) ;
    unsigned tag = gettag(old) ;
    SYNC() ;
    q->next = realold ;                        
    SYNC() ;
    q_packed = pack(q,tag) ;
  } while (!CAS(&p->top,old,q_packed)) ; // Do push
  SYNC() ;
  if (dbg) {
    fprintf(stderr,"PUSH: 0x%"PRIXMAX "\n",q_packed) ;
  }
}

void *pop(stack_t *p,int id) {
  for ( ; ; ) {
    SYNC() ;
    tagged_t old = p->top ; // Old top node.
    SYNC() ;
    stack_node_t *realold = getptr(old) ;
    unsigned tag = gettag(old) ;
    if (realold == NULL) {
      return NULL ;
    } else {
      stack_node_t *realnext = realold->next ;
      SYNC() ;
      tagged_t next = pack(realnext,tag+1) ;
      if (CAS(&p->top,old,next)) { // Do pop
        SYNC() ;
        if (dbg) fprintf(stderr,"POP: 0x%" PRIXMAX "\n",old) ;
        void *r = realold->val ;         // Clean up and return
        free_stack_node(realold) ;
        return r ;
      }
    }
  }
}
